from django.contrib.auth.models import User
from django.db import models

from checklist.models import Checklist


class Project(models.Model):

    name = models.SlugField(max_length=128, unique=True)
    deployed_revision = models.ForeignKey(
        'Revision', blank=True, null=True, related_name="+")
    url = models.CharField('URL', max_length=1024)
    git_branch = models.CharField(
        'git branch (leave empty for bzr)',
        max_length=512, blank=True, null=True)
    sourcedeps_path = models.CharField(max_length=512, blank=True)
    no_qa_of_main_branch = models.BooleanField(default=False)
    description = models.TextField(blank=True, null=True)
    deployment_checklist = models.ForeignKey(Checklist, blank=True, null=True)

    class Meta:
        ordering = ['name']

    def __unicode__(self):
        return u"Project %s" % self.name

    class Factory:

        @staticmethod
        def get_name(field, factory):
            return "project%d" % factory.getUniqueInteger()


class RevisionManager(models.Manager):

    def get_queryset(self):
        """ Fetch the related project by default.

        Since the __unicode__ method needs the project and calling __unicode__
        on a list of Revisions is quite common, this gives a good performance
        boost and reduction of number of queries.
        """
        default_qs = super(RevisionManager, self).get_queryset()
        return default_qs.select_related('project')


class Revision(models.Model):

    (STATUS_UNKNOWN, STATUS_OK, STATUS_BAD, STATUS_UNTESTABLE) = range(4)

    STATUS_CHOICES = (
        (STATUS_UNKNOWN, 'Unknown'),
        (STATUS_OK, 'Yes'),
        (STATUS_BAD, 'No'),
        (STATUS_UNTESTABLE, 'Untestable'),
        )

    objects = RevisionManager()
    project = models.ForeignKey(Project)
    parent = models.ForeignKey(
        'self', blank=True, null=True, related_name="children")
    revno = models.IntegerField(db_index=True)
    commit_sha1 = models.CharField(max_length=40, blank=True, null=True)
    # Blank for the main branch, non-blank for child revisions
    branch = models.CharField(max_length=512, blank=True, null=True)
    status = models.IntegerField(choices=STATUS_CHOICES)
    commit_message = models.TextField()
    assignee = models.CharField(max_length=128, blank=True, null=True)
    committer = models.ForeignKey(User, blank=True, null=True)
    link = models.URLField()
    rolled_back_by = models.ForeignKey(
        'self', blank=True, null=True, related_name="rolls_back")

    class Meta:
        ordering = ['project', 'branch', '-revno']
        unique_together = (('project', 'branch', 'revno'),)

    @classmethod
    def get_summary_status_display(cls, summary_status):
        """Return the status display string for the summary state.

        The summary state includes any child revisions.
        """
        # TODO: tests
        return cls.STATUS_CHOICES[summary_status][1]

    def __unicode__(self):
        project = str(self.project)
        if self.branch:
            project += " (%s)" % self.branch
        if self.commit_sha1:
            return u"Commit %s of %s (%s)" % (self.commit_sha1, project,
                                              self.get_status_display())
        else:
            return u"Revision %d of %s (%s)" % (self.revno, project,
                                                self.get_status_display())

    @property
    def title(self):
        if self.commit_sha1:
            return u"Commit %s (%s)" % (self.commit_sha1, self.revno)
        else:
            return u"Revision %d" % self.revno

    @property
    def description(self):
        if self.commit_sha1:
            return u"commit %s (%s)" % (self.commit_sha1, self.revno)
        else:
            return u"revision %d" % self.revno

    @property
    def assignee_link(self):
        if self.assignee:
            return "https://launchpad.net/~%s" % self.assignee
        return None

    class Factory:

        @staticmethod
        def get_branch(field, factory):
            return ""


class DeploymentRequest(models.Model):

    (STATUS_WAITING, STATUS_COMPLETE) = range(2)

    STATUS_CHOICES = (
        (STATUS_WAITING, 'Waiting'),
        (STATUS_COMPLETE, 'Complete'),
        )

    project = models.ForeignKey(
        Project,
        related_name="deployment_request_set")
    start_revision = models.ForeignKey(Revision, blank=True, null=True,
                                       related_name="+")
    deploy_revision = models.ForeignKey(Revision, related_name="+")
    status = models.IntegerField(choices=STATUS_CHOICES,
                                 default=STATUS_WAITING)
    date_created = models.DateTimeField(auto_now_add=True)
    date_completed = models.DateTimeField(blank=True, null=True)
    requester = models.ForeignKey(
        User, related_name="requested_deployment_set")
    completer = models.ForeignKey(
        User, related_name="completed_deployment_set",
        blank=True, null=True)

    class Meta:
        ordering = ['date_created', 'project', 'deploy_revision']
        unique_together = (('project', 'deploy_revision'),)

    def __unicode__(self):
        return u"DeploymentRequest - %s" % self.deploy_revision
