from django.conf import settings
from django.contrib.auth import login
from django.contrib.auth.models import User
from django.core.urlresolvers import reverse
from django.http import HttpRequest
from django.test.client import Client
from django.utils.importlib import import_module

from django_factory import TestCase

import soupmatchers

from test_api_helpers import *
from test_data import *
from test_templates import *
from test_views import *


class TestClient(Client):

    def login_user(self, user):
        """
        Login as specified user, does not depend on auth backend (hopefully)

        This is based on Client.login() with a small hack that does not
        require the call to authenticate()
        """
        if not 'django.contrib.sessions' in settings.INSTALLED_APPS:
            raise AssertionError("Unable to login without django.contrib.sessions in INSTALLED_APPS")
        user.backend = "%s.%s" % ("django.contrib.auth.backends",
                                  "ModelBackend")
        engine = import_module(settings.SESSION_ENGINE)

        # Create a fake request to store login details.
        request = HttpRequest()
        if self.session:
            request.session = self.session
        else:
            request.session = engine.SessionStore()
        login(request, user)

        # Set the cookie to represent the session.
        session_cookie = settings.SESSION_COOKIE_NAME
        self.cookies[session_cookie] = request.session.session_key
        cookie_data = {
            'max-age': None,
            'path': '/',
            'domain': settings.SESSION_COOKIE_DOMAIN,
            'secure': settings.SESSION_COOKIE_SECURE or None,
            'expires': None,
        }
        self.cookies[session_cookie].update(cookie_data)

        # Save the session values.
        request.session.save()


class RevTrackerTestCase(TestCase):

    def setUp(self):
        super(RevTrackerTestCase, self).setUp()
        self.client = TestClient()

    def create_user_and_log_in(self):
        self.user = User.objects.create(username="eve")
        # Create a groups so that the user is allowed in
        self.user.groups.create(name="group")
        self.client.login_user(self.user)


class IndexProjectsList(soupmatchers.Tag):

    def __init__(self, **kwargs):
        kwargs.setdefault('attrs', {})
        kwargs['attrs']['id'] = 'projects'
        super(IndexProjectsList, self).__init__(
            'index projects list', 'div', **kwargs)


class IndexProject(soupmatchers.Tag):

    def __init__(self, **kwargs):
        kwargs.setdefault('attrs', {})
        kwargs['attrs']['class'] = 'project section sixteen columns alpha'
        super(IndexProject, self).__init__(
            'index project', 'div', **kwargs)


class ProjectName(soupmatchers.Tag):

    def __init__(self, **kwargs):
        kwargs.setdefault('attrs', {})
        kwargs['attrs']['class'] = 'project_name six columns alpha main'
        super(ProjectName, self).__init__(
            'project name', 'div', **kwargs)


class ProjectRevision(soupmatchers.Tag):

    def __init__(self, revision, *args, **kwargs):
        kwargs.setdefault('attrs', {})
        kwargs['attrs']['id'] = 'revision_%s_%d' % (revision.branch, revision.revno)
        if 'class' not in kwargs['attrs']:
            kwargs['attrs']['class'] = "revision %s section sixteen columns alpha" % revision.get_status_display()
        super(ProjectRevision, self).__init__('revision %d' % revision.revno, 'div', *args, **kwargs)


class ProjectDeployedRevision(soupmatchers.Tag):

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('attrs', {})
        kwargs['attrs']['id'] = 'deployed_revision'
        super(ProjectDeployedRevision, self).__init__('deployed revision', 'div', *args, **kwargs)


class ProjectDeployableRevision(soupmatchers.Tag):

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('attrs', {})
        kwargs['attrs']['id'] = 'deployable_revision'
        super(ProjectDeployableRevision, self).__init__('deployable revision', 'div', *args, **kwargs)


class ProjectBlockingRevision(soupmatchers.Tag):

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('attrs', {})
        kwargs['attrs']['id'] = 'blocking_revision'
        super(ProjectBlockingRevision, self).__init__('blocking revision', 'div', *args, **kwargs)


class ProjectRevisionChangeStatusForm(soupmatchers.Tag):

    def __init__(self, revision, *args, **kwargs):
        kwargs.setdefault('attrs', {})
        kwargs['attrs']['id'] = 'change_revision_status_Yes_%s_%d' % (revision.branch, revision.revno)
        if 'method' not in kwargs['attrs']:
            kwargs['attrs']['method'] = 'post'
        if 'action' not in kwargs['attrs']:
            kwargs['attrs']['action'] = reverse('project_change_revision_status',
                    args=(revision.project.name, revision.pk))
        super(ProjectRevisionChangeStatusForm, self).__init__('revision %d change status form' % revision.revno,
                'form', *args, **kwargs)


class ProjectRevisionSetRollbackForm(soupmatchers.Tag):

    def __init__(self, revision, *args, **kwargs):
        kwargs.setdefault('attrs', {})
        kwargs['attrs']['id'] = 'set_revision_rollback_%s_%d' % (revision.branch, revision.revno)
        if 'method' not in kwargs['attrs']:
            kwargs['attrs']['method'] = 'post'
        if 'action' not in kwargs['attrs']:
            kwargs['attrs']['action'] = reverse('project_set_revision_rollback',
                    args=(revision.project.name, revision.pk))
        super(ProjectRevisionSetRollbackForm, self).__init__('revision %d set rollback form' % revision.revno,
                'form', *args, **kwargs)


class ProjectUpdateDeployedRevisionForm(soupmatchers.Tag):

    def __init__(self, project, *args, **kwargs):
        kwargs.setdefault('attrs', {})
        kwargs['attrs']['id'] = 'update_deployed_revision'
        if 'method' not in kwargs['attrs']:
            kwargs['attrs']['method'] = 'post'
        if 'action' not in kwargs['attrs']:
            kwargs['attrs']['action'] = reverse('project_view', args=(project.name,))
        super(ProjectUpdateDeployedRevisionForm, self).__init__('update deployed revision form',
                'form', *args, **kwargs)


