from email.Utils import parseaddr
import os.path
import shutil
import tempfile
from test.test_support import captured_stdout
from urllib import pathname2url

from bzrlib.controldir import ControlDir
from bzrlib.tests import TestCaseWithTransport
from django.test.utils import override_settings
from django_factory import TestCase
import pygit2
from testscenarios import WithScenarios
from testtools.matchers import (
    MatchesSetwise,
    MatchesStructure,
    )

from revtracker import tasks
from revtracker.launchpad import (
    get_launchpad,
    LaunchpadPersonDouble,
    )
from revtracker.models import (
    Project,
    Revision,
    )
from revtracker.tasks import VCS


@override_settings(LP_INSTANCE='fake')
class TasksTestCase(WithScenarios, TestCase, TestCaseWithTransport):

    scenarios = [
        ('bzr', {'vcs': VCS.VCS_BAZAAR}),
        ('git', {'vcs': VCS.VCS_GIT}),
        ]

    def make_commits(self, branch_name='source', git_branch='master',
                     num_commits=1, path='a', contents=None, committer=None):
        if committer is None:
            committer = 'Test User <test-user@example.com>'
        if self.vcs == VCS.VCS_BAZAAR:
            if os.path.exists(branch_name):
                control, _ = ControlDir.open_containing(branch_name)
                tree = control.open_workingtree()
                tree.branch.lock_read()
                try:
                    start_revno = tree.branch.last_revision_info()[0]
                finally:
                    tree.branch.unlock()
            else:
                tree = self.make_branch_and_tree(branch_name)
                start_revno = 0
            for i in range(start_revno + 1, start_revno + num_commits + 1):
                if contents is None:
                    contents = '%d\n' % i
                self.build_tree_contents(
                    [(os.path.join(branch_name, path), contents)])
                tree.add(path)
                tree.commit(message='commit %d' % i, committer=committer)
            return tree.branch, self.get_readonly_url(branch_name)
        else:
            if not git_branch.startswith('refs/heads/'):
                git_branch = 'refs/heads/' + git_branch
            repository_path = os.path.abspath(branch_name)
            repository_url = 'file://' + pathname2url(repository_path)
            if os.path.exists(branch_name):
                repository = pygit2.Repository(branch_name)
                start_revision = tasks.LocalBranch(
                    VCS.VCS_GIT, repository, repository_url).tip_revision()
                start_revno = start_revision.revno
                parents = [repository[start_revision.revid].id]
            else:
                repository = pygit2.init_repository(
                    branch_name, initial_head=git_branch)
                start_revno = 0
                parents = []
            for i in range(start_revno + 1, start_revno + num_commits + 1):
                if contents is None:
                    contents = '%d\n' % i
                with open(os.path.join(branch_name, path), 'w') as f:
                    f.write(contents)
                repository.index.add(path)
                tree = repository.index.write_tree()
                author = pygit2.Signature(*parseaddr(committer))
                commit = repository.create_commit(
                    git_branch, author, author, 'commit %d' % i, tree, parents)
                parents = [commit]
            return repository, repository_url


class GetOrCreateBranchTests(TasksTestCase):

    def test_create_branch(self):
        branch, url = self.make_commits()
        local_branch = tasks.LocalBranch.get_or_create(
            self.vcs, url, 'project')
        try:
            if self.vcs == VCS.VCS_BAZAAR:
                expected_revno, expected_revid = branch.last_revision_info()
            else:
                expected_revno = 1
                expected_revid = branch.revparse_single('HEAD').id.hex
            self.assertThat(
                local_branch.tip_revision(), MatchesStructure.byEquality(
                    revid=expected_revid, revno=expected_revno))
        finally:
            local_branch.unlock()

    def test_get_branch(self):
        branch, url = self.make_commits()
        local_branch = tasks.LocalBranch.get_or_create(
            self.vcs, url, 'project')
        try:
            if self.vcs == VCS.VCS_BAZAAR:
                expected_revno, expected_revid = branch.last_revision_info()
            else:
                expected_revno = 1
                expected_revid = branch.revparse_single('HEAD').id.hex
            self.assertThat(
                local_branch.tip_revision(), MatchesStructure.byEquality(
                    revid=expected_revid, revno=expected_revno))
        finally:
            local_branch.unlock()
        self.make_commits()
        local_branch = tasks.LocalBranch.get_or_create(
            self.vcs, url, 'project')
        try:
            if self.vcs == VCS.VCS_BAZAAR:
                expected_revno, expected_revid = branch.last_revision_info()
            else:
                expected_revno = 2
                expected_revid = branch.revparse_single('HEAD').id.hex
            self.assertThat(
                local_branch.tip_revision(), MatchesStructure.byEquality(
                    revid=expected_revid, revno=expected_revno))
        finally:
            local_branch.unlock()


class GetRevisionLinkTests(TestCase):

    def make_fake_bzr_branch(self):
        class FakeBazaarBranch(object):
            def is_locked(self):
                return True

        return FakeBazaarBranch()

    def test_bzr_https(self):
        branch = tasks.LocalBranch(
            VCS.VCS_BAZAAR, self.make_fake_bzr_branch(),
            'https://bazaar.launchpad.net/+branch/foo')
        self.assertEqual(
            'https://bazaar.launchpad.net/+branch/foo/revision/42',
            branch.get_revision_link(tasks.LocalRevision(None, None, 42)))

    def test_bzr_lp(self):
        branch = tasks.LocalBranch(
            VCS.VCS_BAZAAR, self.make_fake_bzr_branch(), 'lp:foo')
        self.assertEqual(
            'https://bazaar.launchpad.net/+branch/foo/revision/42',
            branch.get_revision_link(tasks.LocalRevision(None, None, 42)))

    def test_git_https(self):
        branch = tasks.LocalBranch(
            VCS.VCS_GIT, None, 'https://git.launchpad.net/foo')
        self.assertEqual(
            'https://git.launchpad.net/foo/commit?id=dummy',
            branch.get_revision_link(tasks.LocalRevision(None, 'dummy', None)))

    @override_settings(
        GIT_SSH_DEFAULT_USERNAME='test-user', GIT_SSH_DEFAULT_PORT=9422)
    def test_git_ssh(self):
        branch = tasks.LocalBranch(
            VCS.VCS_GIT, None,
            tasks.fixup_branch_location('git+ssh://git.launchpad.net/foo'))
        self.assertEqual(
            'https://git.launchpad.net/foo/commit?id=dummy',
            branch.get_revision_link(tasks.LocalRevision(None, 'dummy', None)))

    def test_git_lp(self):
        branch = tasks.LocalBranch(VCS.VCS_GIT, None, 'lp:foo')
        self.assertEqual(
            'https://git.launchpad.net/foo/commit?id=dummy',
            branch.get_revision_link(tasks.LocalRevision(None, 'dummy', None)))


class IterRevisionsTests(TasksTestCase):

    def test_start_is_tip(self):
        branch, url = self.make_commits()
        local_branch = tasks.LocalBranch.get_or_create(
            self.vcs, url, 'project')
        try:
            tip_revision = local_branch.tip_revision()
            self.assertEqual(
                [],
                list(local_branch.iter_revisions(start_revision=tip_revision)))
        finally:
            local_branch.unlock()

    def test_all(self):
        branch, url = self.make_commits(num_commits=2)
        local_branch = tasks.LocalBranch.get_or_create(
            self.vcs, url, 'project')
        try:
            self.assertEqual(
                [2, 1],
                [revision.revno for revision in local_branch.iter_revisions()])
        finally:
            local_branch.unlock()

    def test_bounded(self):
        branch, url = self.make_commits(num_commits=4)
        local_branch = tasks.LocalBranch.get_or_create(
            self.vcs, url, 'project')
        try:
            all_revisions = list(reversed(list(local_branch.iter_revisions())))
            old_revision = all_revisions[0]
            self.assertEqual(1, old_revision.revno)
            new_revision = all_revisions[2]
            self.assertEqual(3, new_revision.revno)
            self.assertEqual(
                [3, 2],
                [revision.revno for revision in local_branch.iter_revisions(
                    start_revision=old_revision, end_revision=new_revision)])
        finally:
            local_branch.unlock()

    def test_tip_only(self):
        branch, url = self.make_commits(num_commits=2)
        local_branch = tasks.LocalBranch.get_or_create(
            self.vcs, url, 'project')
        try:
            tip_revision = local_branch.tip_revision()
            self.assertEqual(
                [2],
                [revision.revno for revision in local_branch.iter_revisions(
                    end_revision=tip_revision, tip_only=True)])
        finally:
            local_branch.unlock()


class AddRevisionToDbTests(TasksTestCase):

    def setUp(self):
        super(AddRevisionToDbTests, self).setUp()
        launchpadlib_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, launchpadlib_dir)
        self.lp = get_launchpad(launchpadlib_dir)

    def test_email_in_map(self):
        project = self.factory.make_one(Project)
        _, url = self.make_commits(committer='Test User <user@example.com>')
        branch = tasks.LocalBranch.get_or_create(self.vcs, url, 'project')
        try:
            tip_revision = branch.tip_revision()
            test_user = LaunchpadPersonDouble('test-user')
            email_map = {'user@example.com': test_user}
            db_revision = branch.add_revision_to_db(
                self.lp, project, tip_revision, email_map)
            self.assertEqual('commit 1', db_revision.commit_message)
            self.assertEqual('test-user', db_revision.assignee)
            self.assertEqual('test-user', db_revision.committer.username)
            if self.vcs == VCS.VCS_BAZAAR:
                self.assertEqual(url + '/revision/1', db_revision.link)
            else:
                self.assertEqual(
                    url + '/commit?id=%s' % tip_revision.revid,
                    db_revision.link)
        finally:
            branch.unlock()

    def test_email_from_launchpad(self):
        project = self.factory.make_one(Project)
        _, url = self.make_commits(committer='Test User <user@example.com>')
        branch = tasks.LocalBranch.get_or_create(self.vcs, url, 'project')
        try:
            tip_revision = branch.tip_revision()
            test_user = LaunchpadPersonDouble('test-user')
            self.lp.people.emails = {'user@example.com': test_user}
            email_map = {}
            db_revision = branch.add_revision_to_db(
                self.lp, project, tip_revision, email_map)
            self.assertEqual('commit 1', db_revision.commit_message)
            self.assertEqual('test-user', db_revision.assignee)
            self.assertEqual('test-user', db_revision.committer.username)
            if self.vcs == VCS.VCS_BAZAAR:
                self.assertEqual(url + '/revision/1', db_revision.link)
            else:
                self.assertEqual(
                    url + '/commit?id=%s' % tip_revision.revid,
                    db_revision.link)
            self.assertEqual({'user@example.com': test_user}, email_map)
        finally:
            branch.unlock()

    def test_bad_email(self):
        project = self.factory.make_one(Project)
        _, url = self.make_commits(committer='Test User <user>')
        branch = tasks.LocalBranch.get_or_create(self.vcs, url, 'project')
        try:
            tip_revision = branch.tip_revision()
            email_map = {}
            db_revision = branch.add_revision_to_db(
                self.lp, project, tip_revision, email_map)
            self.assertEqual('commit 1', db_revision.commit_message)
            self.assertIsNone(db_revision.assignee)
            self.assertIsNone(db_revision.committer)
            if self.vcs == VCS.VCS_BAZAAR:
                self.assertEqual(url + '/revision/1', db_revision.link)
            else:
                self.assertEqual(
                    url + '/commit?id=%s' % tip_revision.revid,
                    db_revision.link)
            self.assertEqual({'user': None}, email_map)
        finally:
            branch.unlock()

    def test_no_parent(self):
        project = self.factory.make_one(Project)
        _, url = self.make_commits(committer='Test User <user>')
        branch = tasks.LocalBranch.get_or_create(self.vcs, url, 'project')
        try:
            tip_revision = branch.tip_revision()
            db_revision = branch.add_revision_to_db(
                self.lp, project, tip_revision, {})
            self.assertIsNone(db_revision.parent)
            self.assertEqual('', db_revision.branch)
        finally:
            branch.unlock()

    def test_parent(self):
        project = self.factory.make_one(Project)
        parent_revision = self.factory.make_one(
            Revision, project=project, revno=1, parent=None, branch='',
            status=Revision.STATUS_UNKNOWN)
        _, url = self.make_commits(committer='Test User <user>')
        branch = tasks.LocalBranch.get_or_create(self.vcs, url, 'project')
        try:
            tip_revision = branch.tip_revision()
            db_revision = branch.add_revision_to_db(
                self.lp, project, tip_revision, {},
                parent_revision=parent_revision)
            self.assertEqual(parent_revision, db_revision.parent)
            self.assertEqual(url, db_revision.branch)
        finally:
            branch.unlock()

    def test_created_qa(self):
        project = self.factory.make_one(Project)
        _, url = self.make_commits(committer='Test User <user>')
        branch = tasks.LocalBranch.get_or_create(self.vcs, url, 'project')
        try:
            tip_revision = branch.tip_revision()
            db_revision = branch.add_revision_to_db(
                self.lp, project, tip_revision, {})
            self.assertEqual(Revision.STATUS_UNKNOWN, db_revision.status)
        finally:
            branch.unlock()

    def test_created_no_qa(self):
        project = self.factory.make_one(Project)
        _, url = self.make_commits(committer='Test User <user>')
        branch = tasks.LocalBranch.get_or_create(self.vcs, url, 'project')
        try:
            tip_revision = branch.tip_revision()
            db_revision = branch.add_revision_to_db(
                self.lp, project, tip_revision, {}, require_qa=False)
            self.assertEqual(Revision.STATUS_OK, db_revision.status)
        finally:
            branch.unlock()


class FixupBranchLocationTests(TestCase):

    def test_lp_scheme(self):
        self.assertEqual(
            'lp:~foo/bar/baz', tasks.fixup_branch_location('lp:~foo/bar/baz'))

    def test_bazaar_launchpad_net(self):
        self.assertEqual(
            'lp:~foo/bar/baz',
            tasks.fixup_branch_location(
                'https://bazaar.launchpad.net/~foo/bar/baz'))

    def test_bazaar_ubunet(self):
        self.assertEqual(
            'lp:~foo/bar/baz',
            tasks.fixup_branch_location(
                'bzr+ssh://bazaar.ubunet/~foo/bar/baz'))

    def test_bazaar_isd(self):
        self.assertEqual(
            'lp:~foo/bar/baz',
            tasks.fixup_branch_location('bzr+ssh://bazaar.isd/~foo/bar/baz'))

    def test_local_file_with_git_directory(self):
        tempdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, tempdir)
        os.mkdir(os.path.join(tempdir, '.git'))
        url = 'file://' + tempdir
        self.assertEqual(url, tasks.fixup_branch_location(url))

    def test_local_file_without_git_directory(self):
        tempdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, tempdir)
        url = 'file://' + tempdir
        self.assertEqual(url, tasks.fixup_branch_location(url))

    def test_git_anonymous(self):
        self.assertEqual(
            'git://git.launchpad.net/~foo/bar/baz',
            tasks.fixup_branch_location(
                'git://git.launchpad.net/~foo/bar/baz'))

    @override_settings(GIT_SSH_DEFAULT_USERNAME='test-user')
    def test_git_ssh(self):
        self.assertEqual(
            'git+ssh://test-user@git.launchpad.net/~foo/bar/baz',
            tasks.fixup_branch_location(
                'git+ssh://git.launchpad.net/~foo/bar/baz'))

    @override_settings(
        GIT_SSH_DEFAULT_USERNAME='test-user', GIT_SSH_DEFAULT_PORT=9422)
    def test_git_ssh_override_port(self):
        self.assertEqual(
            'git+ssh://test-user@git.launchpad.net:9422/~foo/bar/baz',
            tasks.fixup_branch_location(
                'git+ssh://git.launchpad.net/~foo/bar/baz'))

    @override_settings(
        GIT_SSH_DEFAULT_USERNAME='test-user', GIT_SSH_DEFAULT_PORT=9422)
    def test_git_ssh_specific_port(self):
        self.assertEqual(
            'git+ssh://test-user@git.example.net:22/~foo/bar/baz',
            tasks.fixup_branch_location(
                'git+ssh://git.example.net:22/~foo/bar/baz'))

    def test_git_https(self):
        self.assertEqual(
            'https://git.launchpad.net/~foo/bar/baz',
            tasks.fixup_branch_location(
                'git+https://git.launchpad.net/~foo/bar/baz'))


class GuessVCSTests(TestCase):

    def test_git_https(self):
        self.assertEqual(
            VCS.VCS_GIT, tasks.guess_vcs('git+https://git.launchpad.net/foo'))

    def test_git_ssh(self):
        self.assertEqual(
            VCS.VCS_GIT, tasks.guess_vcs('git+ssh://git.launchpad.net/foo'))

    def test_git(self):
        self.assertEqual(
            VCS.VCS_GIT, tasks.guess_vcs('git://git.launchpad.net/foo'))

    def test_bzr_https(self):
        self.assertEqual(
            VCS.VCS_BAZAAR,
            tasks.guess_vcs('bzr+https://bazaar.launchpad.net/foo'))

    def test_bzr_ssh(self):
        self.assertEqual(
            VCS.VCS_BAZAAR,
            tasks.guess_vcs('bzr+ssh://bazaar.launchpad.net/foo'))

    def test_lp(self):
        self.assertEqual(VCS.VCS_BAZAAR, tasks.guess_vcs('lp:foo'))

    def test_file_with_git_directory(self):
        tempdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, tempdir)
        os.mkdir(os.path.join(tempdir, '.git'))
        self.assertEqual(VCS.VCS_GIT, tasks.guess_vcs('file://' + tempdir))

    def test_file_without_git_directory(self):
        tempdir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, tempdir)
        self.assertEqual(VCS.VCS_BAZAAR, tasks.guess_vcs('file://' + tempdir))

    def test_bogus(self):
        # If we don't understand the URL at all, we treat it as Bazaar.
        self.assertEqual(VCS.VCS_BAZAAR, tasks.guess_vcs('bogus'))


class UpdateRevisionsTests(TasksTestCase):

    def test_do_nothing_if_tip_unchanged(self):
        branch, url = self.make_commits(num_commits=2)
        project = self.factory.make_one(Project, url=url, git_branch='master')
        self.factory.make_one(
            Revision, project=project, revno=1, parent=None, branch='',
            status=Revision.STATUS_BAD, commit_message='nonsense 1')
        tip_revision = self.factory.make_one(
            Revision, project=project, revno=2, parent=None, branch='',
            status=Revision.STATUS_BAD, commit_message='nonsense 2')
        if self.vcs == VCS.VCS_GIT:
            tip_revision.commit_sha1 = branch.revparse_single('HEAD').id.hex
            tip_revision.save()
        with captured_stdout():
            tasks.update_revisions(project.pk)
        self.assertThat(project.revision_set.all(), MatchesSetwise(
            MatchesStructure.byEquality(
                project=project, parent=None, revno=1, branch='',
                status=Revision.STATUS_BAD, commit_message='nonsense 1'),
            MatchesStructure.byEquality(
                project=project, parent=None, revno=2, branch='',
                status=Revision.STATUS_BAD, commit_message='nonsense 2'),
            ))

    def test_correct_status(self):
        branch, url = self.make_commits(num_commits=3)
        project = self.factory.make_one(Project, url=url, git_branch='master')
        start_revision = self.factory.make_one(
            Revision, project=project, revno=1, parent=None, branch='',
            status=Revision.STATUS_BAD)
        if self.vcs == VCS.VCS_GIT:
            start_revision.commit_sha1 = (
                    branch.revparse_single('HEAD~2').id.hex)
            start_revision.save()
        with captured_stdout():
            tasks.update_revisions(project.pk)
        self.assertThat(project.revision_set.all(), MatchesSetwise(
            MatchesStructure.byEquality(
                project=project, parent=None, revno=1, branch='',
                status=Revision.STATUS_BAD),
            MatchesStructure.byEquality(
                project=project, parent=None, revno=2, branch='',
                status=Revision.STATUS_UNKNOWN),
            MatchesStructure.byEquality(
                project=project, parent=None, revno=3, branch='',
                status=Revision.STATUS_UNKNOWN),
            ))

    def test_correct_status_for_sourcedeps(self):
        branch, url = self.make_commits()
        _, dep_url = self.make_commits(branch_name='dep', num_commits=3)
        if self.vcs == VCS.VCS_BAZAAR:
            dep_rev1 = '1'
            dep_rev3 = '3'
        else:
            dep_rev1 = 'HEAD~2'
            dep_rev3 = 'HEAD'
        self.make_commits(
            path='config-manager.txt',
            contents='dep %s;revno=%s\n' % (dep_url, dep_rev1))
        self.make_commits(
            path='config-manager.txt',
            contents='dep %s;revno=%s\n' % (dep_url, dep_rev3))
        project = self.factory.make_one(
            Project, url=url, git_branch='master',
            sourcedeps_path='config-manager.txt')
        with captured_stdout():
            tasks.update_revisions(project.pk)
        rev2 = project.revision_set.filter(parent=None, revno=2).get()
        rev3 = project.revision_set.filter(parent=None, revno=3).get()
        self.assertThat(project.revision_set.all(), MatchesSetwise(
            MatchesStructure.byEquality(
                project=project, parent=None, revno=1, branch='',
                status=Revision.STATUS_UNKNOWN),
            MatchesStructure.byEquality(
                project=project, parent=None, revno=2, branch='',
                status=Revision.STATUS_UNKNOWN),
            MatchesStructure.byEquality(
                project=project, parent=rev2, revno=1, branch=dep_url,
                status=Revision.STATUS_UNKNOWN),
            MatchesStructure.byEquality(
                project=project, parent=None, revno=3, branch='',
                status=Revision.STATUS_UNKNOWN),
            MatchesStructure.byEquality(
                project=project, parent=rev3, revno=2, branch=dep_url,
                status=Revision.STATUS_UNKNOWN),
            MatchesStructure.byEquality(
                project=project, parent=rev3, revno=3, branch=dep_url,
                status=Revision.STATUS_UNKNOWN),
            ))

    def test_correct_status_for_changed_sourcedep_branch(self):
        _, url = self.make_commits()
        _, dep1_url = self.make_commits(branch_name='dep1')
        if self.vcs == VCS.VCS_BAZAAR:
            dep1_rev1 = '1'
        else:
            dep1_rev1 = 'HEAD~2'
        self.make_commits(
            path='config-manager.txt',
            contents='dep %s;revno=%s\n' % (dep1_url, dep1_rev1))
        _, dep2_url = self.make_commits(branch_name='dep2', num_commits=3)
        if self.vcs == VCS.VCS_BAZAAR:
            dep2_rev3 = '3'
        else:
            dep2_rev3 = 'HEAD'
        self.make_commits(
            path='config-manager.txt',
            contents='dep %s;revno=%s\n' % (dep2_url, dep2_rev3))
        project = self.factory.make_one(
            Project, url=url, git_branch='master',
            sourcedeps_path='config-manager.txt')
        with captured_stdout():
            tasks.update_revisions(project.pk)
        rev2 = project.revision_set.filter(parent=None, revno=2).get()
        rev3 = project.revision_set.filter(parent=None, revno=3).get()
        self.assertThat(project.revision_set.all(), MatchesSetwise(
            MatchesStructure.byEquality(
                project=project, parent=None, revno=1, branch='',
                status=Revision.STATUS_UNKNOWN),
            MatchesStructure.byEquality(
                project=project, parent=None, revno=2, branch='',
                status=Revision.STATUS_UNKNOWN),
            MatchesStructure.byEquality(
                project=project, parent=rev2, revno=1, branch=dep1_url,
                status=Revision.STATUS_UNKNOWN),
            MatchesStructure.byEquality(
                project=project, parent=None, revno=3, branch='',
                status=Revision.STATUS_UNKNOWN),
            # Since the sourcedep branch has changed, we only consider the
            # tip revision from it.
            MatchesStructure.byEquality(
                project=project, parent=rev3, revno=3, branch=dep2_url,
                status=Revision.STATUS_UNKNOWN),
            ))
