import os
import uuid

import oops_dictconfig
import oops_wsgi
from django.conf import settings
from oops_wsgi import django as oops_wsgi_django
from whitenoise import WhiteNoise

from versioninfo import version_info


class EagerOOPSWSGIHandler(oops_wsgi_django.OOPSWSGIHandler):
    """Pre-generates an oops id on exception."""
    def handle_uncaught_exception(self, request, resolver, exc_info):
        # This will pass the exception information back to oops.  It
        # should not be necessary once
        # https://code.djangoproject.com/ticket/16674 has been merged.
        if 'oops.context' in request.environ:
            request.environ['oops.context']['exc_info'] = exc_info
        # Generate OOPS id early so that we can render the error page right
        # away and not depend on passing thread locals around.
        if 'oops.report' in request.environ:
            unique_id = uuid.uuid4().hex
            request.environ['oops.report']['id'] = "OOPS-%s" % unique_id

        return super(EagerOOPSWSGIHandler, self).handle_uncaught_exception(
            request, resolver, exc_info)

    def __call__(self, environ, start_response):
        def start_response_with_exc_info(status, headers, exc_info=None):
            """Custom start_response callback for wsgi."""
            # This will pass the exception information back to oops_wgi.  It
            # should not be necessary once
            # https://code.djangoproject.com/ticket/16674 has been merged.
            if exc_info is None:
                exc_info = environ['oops.context'].get('exc_info', None)
            return start_response(status, headers, exc_info)
        return super(EagerOOPSWSGIHandler, self).__call__(
            environ, start_response_with_exc_info)


class BzrRevnoMiddleware(object):
    REVNO_HEADER = ('X-Bzr-Revision-Number', version_info['revno'])

    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):

        def custom_start_response(status, headers, exc_info=None):
            headers.append(self.REVNO_HEADER)
            return start_response(status, headers, exc_info)

        return self.app(environ, custom_start_response)


def make_app():
    """Encapsulate our RNR handler in an OOPS app for error reporting."""

    non_oops_app = EagerOOPSWSGIHandler()

    non_oops_app = BzrRevnoMiddleware(non_oops_app)

    static_root = os.path.abspath(os.path.join(
        settings.BASE_DIR, settings.STATIC_ROOT))
    non_oops_app = WhiteNoise(non_oops_app, root=static_root, prefix='static')

    config = oops_dictconfig.config_from_dict(settings.OOPSES)
    oops_wsgi.install_hooks(config)
    soft_timeout = settings.SOFT_TIMEOUT_MILLISECONDS
    oops_app = oops_wsgi.make_app(non_oops_app, config, oops_on_status=['500'],
                                  soft_start_timeout=soft_timeout)
    return oops_app


application = make_app()
